import inspect
import json
from dataclasses import asdict, dataclass, field
from whiffle_client.wind.components.base import BaseComponent


@dataclass
class TurbineModelSpecs(BaseComponent):
    name: str = field(default=None)
    hub_height: float = field(default=None)
    rotor_diameter: float = field(default=None)
    rated_power: float = field(default=None)
    reference_density: float = field(default=None)
    reference_turbulence_intensity: float = field(default=None)
    reference_windspeed: list[float] = field(default=None)
    thrust_coefficient: list[float] = field(default=None)
    power: list[float] = field(default=None)
    public: bool = field(default_factory=bool)



import json
from dataclasses import InitVar, asdict, dataclass, field
from pathlib import Path

import pandas as pd


def csv_loader_constructor(base_dir=None):
    def csv_loader(loader, node):
        path = loader.construct_scalar(node)
        if base_dir is not None:
            path = f"{Path(base_dir).resolve()}/{path}"
        return asdict(CSVLoader(path_to_csv=path))

    return csv_loader


@dataclass
class CSVLoader:
    # Init attributes
    path_to_csv: InitVar[str] = ""

    # Postinit attributes
    name: str = field(default=None)
    location: list[list[float]] = field(default=None)
    turbine_model_name: str = field(default=None)
    include_in_les: bool = field(default=True)
    thrust: bool = field(default=True)

    def __post_init__(self, path_to_csv: str = ""):
        self.data = pd.read_csv(path_to_csv).to_dict(orient="list")
        self.name = self.data["name"]
        self.location = list(zip(self.data["lon"], self.data["lat"]))
        self.turbine_model_name = self.data["type"][0]

    def __repr__(self):
        return_data = {
            "name": self.name,
            "location": self.location,
        }
        if self.turbine_model_name != "metmast":
            return_data.update({"turbine_model_name": self.turbine_model_name})

        return json.dumps(return_data)

    def __getitem__(self, item):
        return getattr(self, item)



from dataclasses import dataclass, field

from whiffle_client.wind.components.turbine import Turbine


@dataclass
class Windfarm:
    name: str = field(default=None)
    turbines: list[Turbine] = field(default=None)
    include_in_les: bool = field(default=True)
    thrust: bool = field(default=True)
    turbine_model_name: str = field(default=None)

    def __post_init__(self):
        are_all_turbine_dict = all([isinstance(t, dict) for t in self.turbines])
        if are_all_turbine_dict:
            self.turbines = [Turbine(**t) for t in self.turbines]
        are_all_turbine_turbine_obj = all(
            [isinstance(t, Turbine) for t in self.turbines]
        )
        if not are_all_turbine_turbine_obj:
            raise ValueError("All turbines must be of type Turbine or dict")



from whiffle_client.io import load_yaml_with_include
from tests.conftest import task_with_yaml_include_path, wind_simulation_task


def test_load_yaml_with_include():
    result = load_yaml_with_include(task_with_yaml_include_path)
    assert isinstance(result["simulations"]["001"]["metmasts"], dict)


def test_load_yaml_with_include_relative_true():
    result = load_yaml_with_include(wind_simulation_task, relative_to_file=True)
    assert isinstance(result["windfarms"]["windfarm_b"], dict)
    assert isinstance(result["windfarms"]["windfarm_c"], dict)



import subprocess

import yaml
from click.testing import CliRunner

from tests.conftest import mock_new_task_response, default_task_path
from whiffle_client import Client
from whiffle_client.entrypoints import whiffle

runner = CliRunner()


class TestCli:
    def test_list_config(self):
        config = Client.get_config()
        res = runner.invoke(whiffle, ["config", "list"])
        assert yaml.safe_dump(config) in res.output

    def test_edit_config(self):
        token = "test_token"
        res = runner.invoke(whiffle, ["config", "edit", "user.token", token])
        assert token in res.output
        res = runner.invoke(whiffle, ["config", "list"])
        assert token in res.output

    def test_run_task(self, mocker):
        mock_new_task_response(mocker)
        mock = mocker.patch("whiffle_client.Client.process")
        runner.invoke(whiffle, ["task", "run", default_task_path])
        mock.assert_called_once()

    def test_task_list(self, mocker):
        mock_new_task_response(mocker)
        mock = mocker.patch("whiffle_client.Client.get_tasks")
        runner.invoke(whiffle, ["task", "list"])
        mock.assert_called_once()

    def test_download_task(self, mocker):
        mock_new_task_response(mocker)
        mock = mocker.patch("whiffle_client.Client.download")
        res = runner.invoke(whiffle, ["task", "download", "123"])
        assert not "Error" in res.output, res.output
        mock.assert_called_once()

    def test_attach_task(self, mocker):
        mock_new_task_response(mocker)
        mock = mocker.patch("whiffle_client.Client.communicate")
        runner.invoke(whiffle, ["task", "attach", "123"])
        mock.assert_called_once()

    def test_cancel_task(self, mocker):
        mock_new_task_response(mocker)
        mock = mocker.patch("whiffle_client.Client.cancel")
        res = runner.invoke(whiffle, ["task", "cancel", "123"])
        assert not "Error" in res.output, res.output
        mock.assert_called_once()

    def test_version(self):
        git_version = subprocess.check_output(["git", "describe"]).decode()
        res = runner.invoke(whiffle, ["--version"])
        # res.output = "whiffle-client, version 0.2.9.post1+git.adbca4e1.dirty"
        cli_version = ".".join(res.output.split(" version ")[-1].split(".")[:3])

        print(res.output, cli_version, res.output.split(" ")[2][:5])
        print(git_version, git_version.split("-")[0], git_version[:5])

        assert cli_version == git_version.split("-")[0]



import os
from pathlib import Path
from typing import Union
from urllib.parse import urlparse


from whiffle_client.decorators import load_data
from whiffle_client.io import stream_download_write_chunks, VALID_DOWNLOAD_FILES
from whiffle_client.wind.mapping.base import BaseMapping
from whiffle_client.wind.components.wind_simulation_task import WindSimulationTask


class WindSimulationTaskEndpoints(BaseMapping):

    URL = "/api/v1/wind"
    RESOURCE_TYPE = "WindSimulationTask"

    @load_data(WindSimulationTask)
    def add(
        self,
        data: Union[str, dict, Path, WindSimulationTask] = None,
    ) -> WindSimulationTask:
        """Add new Wind Simulation Task

        Parameters
        ----------
        data : Union[str, dict, Path, WindSimulationTask], optional
            Either a path to a yaml, the data itself as a dictionary or a WindSimulationTask object
            containing the parameters that define the wind simulation task, by default None

        Returns
        -------
        WindSimulationTask
            Object instance of the wind simulation
        """
        wind_simulation_task = self.session.post_request(
            f"{self.session.server_url}{self.URL}", data=data
        )
        return WindSimulationTask.from_dict(wind_simulation_task.json())

    @load_data(WindSimulationTask)
    def edit(
        self,
        wind_simulation_task_id: str,
        data: Union[str, dict, Path, WindSimulationTask] = None,
    ) -> WindSimulationTask:
        """Edit existing Wind Simulation Task

        Parameters
        ----------
        wind_simulation_task_id : str
            Id of wind simulation task
        data : Union[str, dict, Path, WindSimulationTask], optional
            Either a path to a yaml, the data itself as a dictionary or a WindSimulationTask object
            containing the parameters that define the wind simulation task, by default None

        Returns
        -------
        WindSimulationTask
            Object instance of the wind simulation
        """
        request = self.session.put_request(
            f"{self.session.server_url}{self.URL}/{wind_simulation_task_id}",
            data=data,
        )
        return WindSimulationTask.from_dict(request.json())

    def submit(self, wind_simulation_task_id: str) -> WindSimulationTask:
        """Submit Wind Simulation Task

        Parameters
        ----------
        wind_simulation_task_id : str
            Id of wind simulation task

        Returns
        -------
        WindSimulationTask
            Object instance of the wind simulation
        """
        request = self.session.post_request(
            f"{self.session.server_url}{self.URL}/{wind_simulation_task_id}/submit",
        )
        return WindSimulationTask.from_dict(request.json())

    # Wind simulation commands
    def get_all(self) -> list[WindSimulationTask]:
        """Get a list of all the Wind Simulation Tasks available to the user

        Returns
        -------
        list[WindSimulationTask]
            List of WindSimulationTasks object instances
        """
        request = self.session.get_request(f"{self.session.server_url}{self.URL}")
        return [
            WindSimulationTask.from_dict(wind_simulation_tasks_params)
            for wind_simulation_tasks_params in request.json()
        ]

    def get(self, wind_simulation_task_id: str) -> WindSimulationTask:
        """Get a Wind Simulation Task

        Parameters
        ----------
        wind_simulation_task_id : str
            Id of wind simulation

        Returns
        -------
        WindSimulationTask
            Requested WindSimulationTask
        """
        return WindSimulationTask.from_dict(
            self.session.get_request(
                f"{self.session.server_url}{self.URL}/{wind_simulation_task_id}"
            ).json()
        )

    def delete(self, wind_simulation_task_id: str):
        """Delete Wind Simulation

        NOTE: Just Wind Simulation Tasks in 'drafted' status can be deleted

        Parameters
        ----------
        wind_simulation_task_id : str
            Id of wind simulation

        Returns
        -------
        dict
            Dictionary with the deleted WindSimulationTask id
        """
        return self.session.delete_request(
            f"{self.session.server_url}{self.URL}/{wind_simulation_task_id}"
        ).json()

    def download(
        self,
        wind_simulation_task_id: str,
        file: str = "results",
        output_name: str = None,
        output_dir: str = None,
    ):
        """Download Wind Simulation Task results

        Parameters
        ----------
        wind_simulation_task_id : str
            Id of wind simulation
        file : str, optional
            File of the results files to download, by default "results"
        output_name : str, optional
            Output file name to use if provided, else use same name as remote file, by default None
        output_dir : str, optional
            Output directory path, by default None

        Raises
        ------
        ValueError
            Raises error if file type is not in the valid file types
        """
        if not output_dir:
            output_dir = os.getcwd()
        if file and file not in VALID_DOWNLOAD_FILES:
            err_msg = f"File type <{file}> is not in the valid file types <{VALID_DOWNLOAD_FILES}>"
            raise ValueError(err_msg)

        download_url = f"{self.session.server_url}{self.URL}/{wind_simulation_task_id}/download_url"
        download_url += f"?file_path={file}" if file else ""
        wind_simulation_task_file_url = self.session.get_request(download_url).json()[
            "url"
        ]
        # NOTE: If provided, use given name, else take same name as remote
        if output_name:
            local_filename = output_name
        else:
            local_filename = os.path.basename(
                urlparse(wind_simulation_task_file_url).path
            )
        print(
            f"Fetch <{file}> data of simulation <{wind_simulation_task_id}> and store it at: <{local_filename}>"
        )
        stream_download_write_chunks(
            f"{output_dir}/{local_filename}",
            wind_simulation_task_file_url,
            self.session,
        )



from abc import ABC, abstractmethod


class BaseMapping(ABC):
    URL = None
    RESOURCE_TYPE = None

    def __init__(self, session):
        self.session = session

    @abstractmethod
    def add(self):
        raise NotImplementedError

    @abstractmethod
    def edit(self):
        raise NotImplementedError

    @abstractmethod
    def get_all(self):
        raise NotImplementedError

    @abstractmethod
    def get(self):
        raise NotImplementedError

    @abstractmethod
    def delete(self):
        raise NotImplementedError

    @abstractmethod
    def download(self):
        raise NotImplementedError



from dataclasses import asdict, dataclass, field

from whiffle_client.wind.components.base import BaseComponent
from whiffle_client.wind.components.dates import Dates
from whiffle_client.wind.components.domain import Domain
from whiffle_client.wind.components.metmast import Metmast
from whiffle_client.wind.components.turbine import Turbine
from whiffle_client.wind.components.windfarm import Windfarm


@dataclass
class WindSimulationTask(BaseComponent):
    # Metadata attributes
    reference_code: str = field(default=None)
    name: str = field(default=None)

    # Time attributes
    dates: Dates = field(default=None)

    # Domain/object/geolocated elements attributes
    windfarms: list[Windfarm] = field(default_factory=list)
    metmasts: list[Metmast] = field(default_factory=list)
    metmasts_heights: list[float] = field(default_factory=list)
    additional_area: Domain = field(default_factory=dict)

    # Output attributes
    wind_resource_grid_heights: list[float] = field(default_factory=list)
    fields_heights: list[float] = field(default_factory=list)

    def __post_init__(self):
        if isinstance(self.metmasts, list):
            self.metmasts = [
                Metmast(
                    name=metmast["name"],
                    longitude=metmast["longitude"],
                    latitude=metmast["latitude"],
                )
                for metmast in self.metmasts
            ]
        elif isinstance(self.metmasts, dict):
            self.metmasts = [
                Metmast(name=name, latitude=latitude, longitude=longitude)
                for name, (longitude, latitude) in zip(
                    self.metmasts["name"], self.metmasts["location"]
                )
            ]

        if isinstance(self.windfarms, list):
            self.windfarms = [Windfarm(**windfarm) for windfarm in self.windfarms]
        elif isinstance(self.windfarms, dict):
            self.windfarms = [
                Windfarm(
                    name=windfarm_name,
                    include_in_les=windfarm["include_in_les"],
                    thrust=windfarm["thrust"],
                    turbine_model_name=windfarm["turbine_model_name"],
                    turbines=[
                        Turbine(
                            name=name,
                            longitude=longitude,
                            latitude=latitude,
                        )
                        for name, (longitude, latitude) in zip(
                            windfarm["name"], windfarm["location"]
                        )
                    ],
                )
                for windfarm_name, windfarm in self.windfarms.items()
            ]

    def _get_api_params(self):
        params = asdict(self)

        if self.metmasts:
            params["metmasts"] = [
                {
                    "name": metmast.name,
                    "longitude": metmast.longitude,
                    "latitude": metmast.latitude,
                }
                for metmast in self.metmasts
            ]

        if self.windfarms:
            params["windfarms"] = [
                {
                    "name": windfarm.name,
                    "include_in_les": windfarm.include_in_les,
                    "thrust": windfarm.thrust,
                    "turbine_model_name": windfarm.turbine_model_name,
                    "turbines": [
                        {
                            "name": turbine.name,
                            "longitude": turbine.longitude,
                            "latitude": turbine.latitude,
                        }
                        for turbine in windfarm.turbines
                    ],
                }
                for windfarm in self.windfarms
            ]

        return params



import json
import pkg_resources
import platformdirs
import yaml
import os

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from whiffle_client.decorators import request_ok


class BaseClient:
    """
    Base class client to connect to Whiffle APIs
    """

    # API variables
    ENDPOINTS_URL: str = ""

    CONFIG_FILE_NAME = "whiffle_config.yaml"
    CONFIG_FILE_PATH_LOCATIONS = [
        f"{platformdirs.user_config_dir('whiffle')}/{CONFIG_FILE_NAME}",  # app path
        f"{CONFIG_FILE_NAME}",  # workdir path
        pkg_resources.resource_filename(
            "whiffle_client", f"resources/{CONFIG_FILE_NAME}"
        ),  # package resource path
    ]

    # Type method
    def __init__(self, access_token=None, url=None, session=None):
        """
        Initialize the client.

        Authentication order:
        1. `access_token` passed when creating class.
        2. token in CONFIG_FILE_PATH_LOCATIONS (JSON format)

        Parameters
        ----------
        access_token : str, optional
            Token for client session auth
        url : str, optional
            Url pointing to API
        """

        if access_token is None:
            config = self.get_config()
            access_token = config["user"]["token"]
        if url is None:
            config = self.get_config()
            url = config["whiffle"]["url"]
        self.server_url = url

        if session is None:
            # More docs: https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.Retry
            status_forcelist = (500, 502, 503, 504)
            retry = Retry(
                total=5,  # Total number of retries to allow
                backoff_factor=0.1,  # Incremental time between retry requests
                status_forcelist=status_forcelist,  # A set of integer HTTP status codes that will force a retry on
            )
            adapter = HTTPAdapter(max_retries=retry)
            session = requests.Session()
            session.mount("http://", adapter)
            session.mount("https://", adapter)
        self.session = session

        self.session.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {access_token}",
        }

    def __repr__(self) -> str:
        return f"Whiffle wind client connected to url: {self.server_url}"

    @staticmethod
    def get_config():
        """Gathers client configuration from resources or from user config directory.

        Returns
        -------
        dict
            Dictionary containing the configuration.

        Raises
        ------
        FileNotFoundError
            Raises error if no configuration found
        """
        config = None
        for file_path in BaseClient.CONFIG_FILE_PATH_LOCATIONS:
            try:
                with open(file_path) as file_object:
                    config = yaml.safe_load(file_object)
                if config:
                    break
            except FileNotFoundError:
                continue
        else:
            raise FileNotFoundError(
                f"No valid config found on either of {BaseClient.CONFIG_FILE_PATH_LOCATIONS} locations"
            )

        return config

    @staticmethod
    def set_config(config):
        config_file_path = BaseClient.CONFIG_FILE_PATH_LOCATIONS[0]
        os.makedirs(os.path.dirname(config_file_path), exist_ok=True)
        with open(config_file_path, "w") as file_object:
            yaml.safe_dump(config, file_object)

    @request_ok
    def get_request(self, url, **kwargs):
        return self.session.get(url, **kwargs)

    @request_ok
    def post_request(self, url, data=None):
        return self.session.post(url, data=json.dumps(data))

    @request_ok
    def put_request(self, url, data=None):
        return self.session.put(url, data=json.dumps(data))

    @request_ok
    def delete_request(self, url):
        return self.session.delete(url)



from dataclasses import dataclass, field


@dataclass
class Geometry:
    coordinates: list[list[list[float]]] = field(default=None)
    type: str = field(default="Polygon")


@dataclass
class Domain:
    geometry: Geometry = field(default_factory=dict)
    properties: dict = field(default_factory=dict)
    type: str = field(default="Feature")
